import unittest

from flask.ext.toybox.views import NegotiatingMethodView
from flask.ext.toybox import ToyBox
from flask import Flask, request
import json

class EchoView(NegotiatingMethodView):
    def get(self):
        """Echoes back a list of GET query param pairs."""
        return request.args.items()

    def post(self):
        """Echoes back a received body."""
        return request.decoded_data

class NegotiationTestCase(unittest.TestCase):
    def setUp(self):
        app = Flask(__name__)
        toybox = ToyBox(app)
        app.add_url_rule("/echo", view_func=EchoView.as_view("echo"))
        self.app = app.test_client()

    def test_unacceptable(self):
        response = self.app.get("/echo", headers={"Accept": "text/x-spam"})
        self.assertEqual(response.status_code, 406)

    def test_unsupported(self):
        response = self.app.post(
            "/echo", data="spam", content_type="text/x-spam",
            headers={"Accept": "application/json"}
        )
        self.assertEqual(response.status_code, 415)

    def test_echo_get(self):
        response = self.app.get("/echo?spam=sv&eggs=ev", headers={
            "Accept": "application/json",
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            json.loads(response.data),
            [["eggs", "ev"], ["spam", "sv"]]
        )

    def test_echo_post(self):
        data = {"eggs": "ev", "spam": "sv"}
        response = self.app.post(
            "/echo",
            data=json.dumps(data),
            content_type="application/json",
            headers={"Accept": "application/json"}
        )
        self.assertEqual(response.status_code, 200, response.status)
        self.assertEqual(json.loads(response.data), data)


from __future__ import absolute_import

from .exceptions import PreconditionRequired, NotModified
from flask import abort
import base64
import hashlib
import types

METHODS_REQUIRE_IF_MATCH = {"PUT", "DELETE", "PATCH"}

def _raw_object_serialize(obj):
    """
    Unsafe serializer of raw object's content using `repr`.

    This intended to be used internally by `ETagger` class, when no serializer
    is provided.
    """
    d = obj.as_dict(check_permissions=False)
    for k in sorted(d.keys()):
        yield k
        yield ":"
        yield repr(d[k])

class ETagger(object):
    def __init__(self, req, serializer):
        self.etag = None
        self.serializer = serializer
        self.req = req

    def set_etag(self, etag):
        if self.req.method in METHODS_REQUIRE_IF_MATCH:
            # The problem with weak tags, is that If-Match MUST use strong
            # comparsion. This was fixed in HTTPbis, for further details
            # see http://trac.tools.ietf.org/wg/httpbis/trac/ticket/116
            if not self.req.if_match:
                raise PreconditionRequired
            elif etag not in self.req.if_match:
                abort(412)
        elif self.req.method == "GET":
            self.etag = etag
            if self.req.if_none_match and etag in self.req.if_none_match:
                raise NotModified

    def set_raw(self, data, prefix="raw"):
        etag = hashlib.sha1()
        if type(data) is types.GeneratorType:
            for element in data: etag.update(element)
        else:
            etag.update(data)
        etag.update(data)
        digest = base64.b64encode(etag.digest()).rstrip("=")
        self.set_etag("{0}-{1}".format(prefix, digest))

    def set_object(self, obj):
        if self.serializer is not None:
            pname = getattr(self.serializer, "name",
                            self.serializer.__class__.__name__.lower())
            serialize = self.serializer.serialize
        else:
            pname = "none"
            serialize = _raw_object_serialize
        self.set_raw(serialize(obj), pname)


"""
TBD.
"""
from __future__ import absolute_import

from flask import Response, request, current_app, g
from flask.views import MethodView
import werkzeug.exceptions
from . import exceptions, etags
from .utils import is_printable
from functools import wraps

def append_vary(response, vary_on):
    """
    Given a `Response` objects and a header name, checks whenever a header is
    listed in response's `Vary` header(s), and adds one if missing.

    If there are multiple `Vary` headers already, another one is added.
    If there's only single `Vary` header, name is appended to the end.

    Note, the header name comparsion is case insensitive.
    """
    if type(vary_on) in (list, tuple, set, frozenset):
        for header in vary_on:
            append_vary(response, header)
        return

    vary = response.headers.getlist("Vary")
    current_set = set(
        ((name.strip().lower() for name in value.split(",")) for value in vary)
    )
    if vary_on.lower() not in current_set:
        if len(vary) == 1:
            response.headers.set("Vary", vary[0] + ", " + vary_on)
        else:
            response.headers.add("Vary", vary_on)

class NegotiatingMethodView(MethodView):
    """
    A `MethodView`-derived class that negotiates request and response
    content types.

    Two things happen in this class:

    1. If there's a request body, flask's `request` object gets a `decoded_data`
       attribute, that contains Python objects ready for API consumption.

       If request body can't be handled due to unknown or missing `Content-Type`
       header, an `UnsupportedMediaType` exception is raised.

       If there's no request body, the `decoded_data` attribute will be set
       to `None`. If you want to know whenever there was a request body that
       deserialized to `None` or wasn't any, take a look at `request.data`.

    2. In verb-handling methods you should return raw python objects.
       They'll get automatically serialized to a negotiated Content-Type.

       If no content-type could be negotiated due to unacceptable `Accept`
       request header, an `NotAcceptable` exception is raised.
       This negotiation happens *before* the request is handled.

    Example::

        class EchoView(NegotiatingMethodView):
            def get(self):
                # Echoes back a list of GET query param pairs.
                return request.args.items()

            def post(self):
                # Echoes back a received body.
                return request.decoded_data

    To override default (de)serializers you may define two class properties:

    - `SERIALIZERS` - an *ordered* dictionary of serializers, in order of
      preference. Keys are strings of MIME types, values are serializer objects,
      implementing `serialize` method. See the `serialization` package for
      further details.

    - `DESERIALIZERS` - an iterable (preferably, a set) of deserializer classes.

    """
    def negotiate_serializer(self, *args, **kwargs):
        """
        Decide on which serializer will be used to output data.
        Returns a tuple, with first element being MIME type string and
        second being the serializer object to use.

        Raises `NotAcceptable` if no acceptable serializer found, or
        `InternalServerError` if no serializers are defined.
        """
        serializers = getattr(self, "SERIALIZERS",
                              current_app.config["TOYBOX_SERIALIZERS"])

        if len(serializers) > 0:
            mime_type = request.accept_mimetypes.best_match(serializers.keys())
            if mime_type is None:
                raise werkzeug.exceptions.NotAcceptable()
            return mime_type, serializers[mime_type]
        else:
            raise werkzeug.exceptions.InternalServerError()

    def dispatch_request(self, *args, **kwargs):
        if request.method.lower() == "POST":
            method_override = request.headers.get("X-HTTP-Method-Override", None)
            if method_override is not None:
                request.method = method_override

        mime_type, serializer = self.negotiate_serializer(*args, **kwargs)

        # Deserialize the incoming request data (if any)
        deserializers = getattr(self, "DESERIALIZERS",
                                current_app.config["TOYBOX_DESERIALIZERS"])

        decoded_data = None
        has_data = request.data is not None and len(request.data) > 0
        if len(deserializers) > 0 and has_data:
            acceptable = False
            for deserializer in deserializers:
                if request.mimetype in deserializer.mime_types:
                    decoded_data = deserializer.deserialize(request.data)
                    acceptable = True
                    break
            if not acceptable:
                raise werkzeug.exceptions.UnsupportedMediaType()
        else:
            decoded_data = None

        # TODO: Document hydration/dehydration process.
        request.dehydrated_decoded_data = decoded_data
        if hasattr(self, "hydrate"):
            decoded_data = self.hydrate(decoded_data)
        request.decoded_data = decoded_data

        # Provide g.etag_object for ETags
        g.etagger = etagger = etags.ETagger(request, serializer)

        # Call parent.
        result = super(NegotiatingMethodView, self)\
                 .dispatch_request(*args, **kwargs)

        # Handle view's result (response)
        if isinstance(result, Response):
            response = result
        else:
            if type(result) is tuple:
                result, status, headers = result
            else:
                status = None
                headers = None

            if hasattr(self, "dehydrate"):
                result = self.dehydrate(result)

            response = Response(serializer.serialize(result),
                                status, headers, mimetype=mime_type)
            response.serialized_with = serializer

        append_vary(response, ["Accept", "Accept-Encoding"])
        if etagger.etag is not None:
            response.set_etag(etagger.etag)
        return response

class ModelView(NegotiatingMethodView):
    """
    A `NegotiatingMethodView`-based class, that simplifies working with
    database models.

    This class contains pseudo-abstract method `fetch_object` that you
    must implement in derived classes.
    """

    # Can't use @abstractmethod here due to a metaclass conflict, sorry.
    def fetch_object(self, *args, **kwargs):
        """
        Implement this method in your class.

        Return a (non-dehydrated) object, that matches the request.
        If requested object does not exist, raise `NotFound` exception.
        """
        # TODO: Document object interface (get_columns, check_permissions, etc.)
        raise werkzeug.exceptions.NotImplemented( # pragma: no cover
            "<p>This resource is not implemented on server.</p>")

    def get_object(self, *args, **kwargs):
        """
        Returns an object, the same `fetch_object` returns.
        However, this cached object in `cached_object` property, so database
        (or other model data source) is not queried twice.

        Use this instead of `fetch_object` in view's code.

        Beware, for now object is currently cached independently from
        function's arguments, so calling this again with another arguments will
        lead to errors. So, don't.
        """
        if not hasattr(self, "cached_object") or self.cached_object is None:
            self.cached_object = self.fetch_object(*args, **kwargs)
        return self.cached_object

    def get_columns(self, *args, **kwargs):
        return self.get_object(*args, **kwargs).get_columns()

    def get(self, *args, **kwargs):
        obj = self.get_object(*args, **kwargs)
        headers = {}

        if hasattr(obj, "check_permissions"):
            access = obj.check_permissions()
            if len(access) > 0 and access != {"system"}:
                headers["X-Access-Classes"] = ", ".join(sorted(access))

        return obj, 200, headers

    # TODO: Implement `put` method for object creation and updates.
    # TODO: Implement `delete` method for object deletion.

    def patch(self, *args, **kwargs):
        obj = self.get_object(**kwargs)

        if hasattr(obj, "check_permissions"):
            access = obj.check_permissions()
            columns = {c.name: c.permissions.get("writeable", set())
                       for c in self.get_columns(only_db_columns=True)}
        else:
            access = frozenset(["system"])
            columns = {c.name: frozenset(["system"])
                       for c in self.get_columns(only_db_columns=True)}

        r = {}
        for k, v in request.decoded_data.items():
            name = '"{0}"'.format(k) if is_printable(k) else repr(k)

            if k not in columns:
                error = "<p>No such attribute: {0}</p>".format(name)
                raise exceptions.UnprocessableEntity(error)
            if len(columns.get(k, set()) & access) > 0:
                r[k] = v #, list(columns.get(k, set()) & access)
            else:
                error = "<p>Attribute {0} is not writeable.</p>".format(name)
                raise exceptions.UnprocessableEntity(error)

        for k, v in r.items():
            setattr(obj, k, v)
        if hasattr(self, "save_object"):
            self.save_object(obj)

        return Response(status=204)



from __future__ import absolute_import

from .serialization import JSON
from collections import OrderedDict

class ToyBox(object):
    def __init__(self, app):
        self.app = app
        self.init_app(app)

    def init_app(self, app):
        app.config.setdefault("TOYBOX_SERIALIZERS", OrderedDict([
            ("application/json", JSON),
            ("text/json", JSON),
        ]))
        app.config.setdefault("TOYBOX_DESERIALIZERS", {JSON})



import unittest

from flask.ext.toybox import serialization
import collections
import datetime
import decimal

SpamTuple = collections.namedtuple("spam_tuple", "spam, eggs")

class SpamList(object):
    def __init__(self, data):
        self.data = data

    def as_list(self):
        return list(self.data)

class SpamDict(object):
    def __init__(self, data):
        self.data = data

    def as_dict(self):
        return dict(self.data)

class SpamJSON(object):
    def __init__(self, data):
        self.data = data

    def to_json(self):
        return self.data

class SerializationTestCase(unittest.TestCase):
    def test_json_encode_namedtuple(self):
        data = SpamTuple("spam_value", "eggs_value")
        serialized = serialization.JSON.serialize(data)
        reference = '{"spam": "spam_value", "eggs": "eggs_value"}'
        self.assertEqual(serialized, reference)

    def test_json_encode_to_json(self):
        data = SpamJSON("42")
        serialized = serialization.JSON.serialize(data)
        reference = '42'
        self.assertEqual(serialized, reference)

    def test_json_default(self):
        data = collections.OrderedDict([
            ("datetime", datetime.datetime(2012, 1, 1, 0, 0, 0, 0)),
            ("decimal", decimal.Decimal("4242.42424242424242424242")),
            ("as_list", SpamList((1, 2, 3))),
            ("as_dict", SpamDict({"spam": "spam_value", "eggs": "eggs_value"})),
            ("normal", [{"k": "v"}, 123])
        ])

        serialized = serialization.JSON.serialize(data)

        reference = ('{"datetime": "2012-01-01T00:00:00", '
                     '"decimal": "4242.42424242424242424242", '
                     '"as_list": [1, 2, 3], '
                     '"as_dict": {"eggs": "eggs_value", "spam": "spam_value"}, '
                     '"normal": [{"k": "v"}, 123]}')

        self.assertEqual(serialized, reference)



"""
This module describes various helpful exceptions, that may happen
during API lifetime.
"""

from werkzeug.exceptions import HTTPException
from flask import redirect, Response

class UnprocessableEntity(HTTPException):
    """
    The 422 (Unprocessable Entity) status code means the server understands the
    content type of the request entity (hence a 415 (Unsupported Media Type)
    status code is inappropriate), and the syntax of the request entity is
    correct (thus a 400 (Bad Request) status code is inappropriate) but was
    unable to process the contained instructions.

    For example, this error condition may occur if an XML request body contains
    well-formed (i.e., syntactically correct), but semantically erroneous,
    XML instructions.
    """
    code = 422
    description = "<p>Unprocessable Entity.</p>"

class NotModified(HTTPException):
    """
    If the client has performed a conditional GET request and access is
    allowed, but the document has not been modified, the server SHOULD
    respond with this status code. The 304 response MUST NOT contain a
    message-body, and thus is always terminated by the first empty line
    after the header fields.
    """
    code = 304
    def get_response(self, environment):
        return Response(status=304)

class PreconditionRequired(HTTPException):
    """
    The 428 status code indicates that the origin server requires the
    request to be conditional.

    Its typical use is to avoid the "lost update" problem, where a client
    GETs a resource's state, modifies it, and PUTs it back to the server,
    when meanwhile a third party has modified the state on the server,
    leading to a conflict.  By requiring requests to be conditional, the
    server can assure that clients are working with the correct copies.
    """
    code = 428
    description = ('<p>This request is required to be '
                   'conditional; try using "If-Match".</p>')
    name = "Precondition Required"

    def get_response(self, environment):
        resp = super(PreconditionRequired, self).get_response(environment)
        resp.status = str(self.code) + " " + self.name.upper()
        return resp

class Redirect(HTTPException):
    """
    HTTP redirect (30x codes) as exception.
    """
    def __init__(self, location, code=307):
        self.location = location
        self.code = code
        super(Redirect, self).__init__()

    def get_response(self, environment):
        return redirect(self.location, self.code)

def abort_redirect(location, code=307):
    raise Redirect(location, code)



import unittest

from flask.ext.toybox.views import ModelView
from flask.ext.toybox.permissions import make_I, DEFAULT_ACCESS_HIER, ModelColumnInfo
from flask.ext.toybox import ToyBox
from flask import Flask, g
import json

DUMMY_DATA = {
    "spam": "spam_value",
    "eggs": "eggs_value"
}

I = make_I()
class DummyModel(object):
    permissions_test = [
        I("r:authenticated+,w:owner"),
        I("r:all,w:none"),
    ]

    def __init__(self, data):
        self.data = data

    def as_dict(self):
        return self.data

    #def check_permissions(self):
    #    return frozenset([])

    def get_columns(self, **kwargs):
        return {ModelColumnInfo(self, "spam"),
                ModelColumnInfo(self, "eggs")}

class DummyModelView(ModelView):
    def fetch_object(self, *args, **kwargs):
        obj = DummyModel(DUMMY_DATA)
        g.etagger.set_object(obj)
        return obj

class SimpleModelTestCase(unittest.TestCase):
    def setUp(self):
        app = Flask(__name__)
        app.debug = True
        toybox = ToyBox(app)
        app.add_url_rule("/test", view_func=DummyModelView.as_view("test"))
        self.app = app.test_client()

    def test_permission_calc(self):
        reference = [
            {
                "readable": {"authenticated", "owner", "staff",
                             "admin", "system"},
                "writeable": {"owner"},
            },
            {
                "readable": set(DEFAULT_ACCESS_HIER),
                "writeable": set(),
            },
        ]
        self.assertEqual(DummyModel.permissions_test, reference)

    def test_get(self):
        response = self.app.get("/test", headers={"Accept": "application/json"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.data), DUMMY_DATA)
        etag = response.headers.get("ETag", None)
        self.assertIsNotNone(etag)
        response = self.app.get("/test", headers={
            "Accept": "application/json",
            "If-None-Match": etag
        })
        self.assertEqual(response.status_code, 304, response.data)

    def test_patch(self):
        response = self.app.get("/test", headers={"Accept": "application/json"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.data), DUMMY_DATA)
        etag = response.headers.get("ETag", None)
        self.assertIsNotNone(etag)

        response = self.app.patch(
            "/test",
            headers={
                "Accept": "application/json",
                "If-Match": etag
            },
            data=json.dumps({"spam": "Spam Spam Spam"}),
            content_type="application/json"
        )
        self.assertEqual(response.status_code, 204, response.data)


"""
Support for using SQLAlchemy models.

This module features `ModelMixin` class, that you should mix into your models
to ease exposing them with ToyBox.

See `ModelMixin` documentation for details.

Note, if PyYaml is installed, a SafeRepresenter as YAML hashmap is added.
"""

from __future__ import absolute_import

from collections import OrderedDict
from sqlalchemy.orm import column_property, class_mapper, relationship, ColumnProperty, object_session
from sqlalchemy.schema import Column
from .views import ModelView
from .permissions import ModelColumnInfo
from flask import g
from werkzeug.exceptions import InternalServerError

def column_info(model, name, column):
    return ModelColumnInfo(model, name,
                           db_column=isinstance(column, Column),
                           permissions=getattr(column, "info"))

class SAModelMixin(object):
    """
    SQLALchemy model mixin.

    Mix this class to your SQLAlchemy models (not views), to ease using them
    with `ModelView` classes.
    """

    @classmethod
    def _get_permissions(cls, column, what="readable"):
        permissions = column.permissions
        return permissions.get(what, frozenset(["system"]))

    def get_columns(self, only_db_columns=False):
        columns = [
            column_info(self, prop.key, prop.columns[0])
            for prop in class_mapper(self.__class__).iterate_properties
            if isinstance(prop, ColumnProperty) and len(prop.columns) == 1\
               and not prop.key.startswith("_") and\
               (not only_db_columns or isinstance(prop.columns[0], Column))
        ]
        if not only_db_columns:
            # If there's a mix, "real" DB columns should go first
            columns.sort(key=lambda c: c.db_column, reverse=True)
        return columns

    def check_permissions(self):
        return frozenset(["system"])

    def as_dict(self, check_permissions=True):
        columns = self.get_columns()

        if check_permissions:
            levels = self.check_permissions()
            columns = [c for c in columns
                       if any(l in self.__class__._get_permissions(c)
                              for l in levels)]

        return OrderedDict((c.name, getattr(self, c.name)) for c in columns)

    @staticmethod
    def yaml_safe_representer(dumper, data):
        return dumper.represent_mapping(
            u'tag:yaml.org,2002:map', data.as_dict().items(), flow_style=False)

try:
    from yaml.representer import SafeRepresenter
    SafeRepresenter.add_multi_representer(
        SAModelMixin, SAModelMixin.yaml_safe_representer)
    del SafeRepresenter
except ImportError: # pragma: no cover
    pass

def hasUserMixin(owner_id_field):
    """
    Quite specific helper, that was made to ease working with permissions
    for models referencing Django-like user objects.
    """
    class HasUserMixin(object):
        def check_permissions(self, user=None):
            if user is None and hasattr(g, "user"):
                user = g.user

            p = set()
            if user is not None:
                p.add("authenticated")
                if user.is_superuser: p.add("admin")
                if user.is_staff: p.add("staff")
                if user.id == getattr(self, owner_id_field): p.add("owner")
            else:
                p.add("anonymous")
            return p
    return HasUserMixin

def saModelView(session):
    """
    Given an SQLAlchemy session object, returns a class that provides some
    possibly useful default implementations for `get_object`.
    """
    class SAModelView(ModelView):
        def __init__(self, *args, **kwargs):
            if not hasattr(self, "model") or len(args) > 0:
                raise InternalServerError("<p>Server entity misconfiguration.</p>")
            return super(SAModelView, self).__init__(*args, **kwargs)

        def fetch_object(self, *args, **kwargs):
            model = self.model
            obj = session.query(model).filter_by(**kwargs).one()
            if hasattr(g, "etagger"):
                g.etagger.set_object(obj)
            return obj

        def get_columns(self, *args, **kwargs):
            # TODO: SAModelMixin duplicates this! Refactor the code.
            only_db_columns = kwargs.pop("only_db_columns", False)
            columns = [
                column_info(self, prop.key, prop.columns[0])
                for prop in class_mapper(self.model).iterate_properties
                if isinstance(prop, ColumnProperty) and len(prop.columns) == 1\
                   and not prop.key.startswith("_") and\
                   (not only_db_columns or isinstance(prop.columns[0], Column))
            ]
            if not only_db_columns:
                # If there's a mix, "real" DB columns should go first
                columns.sort(key=lambda c: c.db_column, reverse=True)
            return columns

    return SAModelView



from itertools import dropwhile
from functools import partial

DEFAULT_ACCESS_HIER = ["anonymous", "authenticated", "owner",
                       "staff", "admin", "system"]
DEFAULT_ACCESS_TARGETS = {"r": "readable", "w": "writeable"}

def I(access_hier, access_targets, access, **kwargs):
    """
    Consider obtaining a partial using convenience helper function `make_I`
    instead of calling this directly from your code.

    Note `"all"` and `"none"` are reserved level names that can't be used in
    `access_hier` list. Consider using synonyms like `"everyone"` or `"nobody"`,
    if you really need them.

    >>> I(DEFAULT_ACCESS_HIER, DEFAULT_ACCESS_TARGETS, "rw:owner,r:staff+")
    {"readable": set(["owner", "staff", "admin", "system"]),
     "writeable": set["owner"]}

    >>> I(DEFAULT_ACCESS_HIER, DEFAULT_ACCESS_TARGETS, "r:staff+")
    {"readable": set(["staff", "admin", "system"]),
     "writeable": set()}

    >>> I(DEFAULT_ACCESS_HIER, DEFAULT_ACCESS_TARGETS, "r:owner+,w:staff+")
    {"readable": set(["owner", "staff", "admin", "system"]),
     "writeable": set(["staff", "admin", "system"])}
    """
    # TODO: Properly document `I` function.
    info = kwargs.copy()
    info.update({"readable": set(), "writeable": set()})

    for name in access.split(","):
        name = name.strip()

        targets = "rw"
        if ":" in name:
            targets, name = name.split(":", 1)

        if name == "all":
            levels = set(access_hier)
        elif name == "none":
            levels = set([])
        elif name.endswith("+"):
            name = name[:-1]
            levels = set(dropwhile(lambda n: name != n, access_hier))
        else:
            levels = {name}

        for target in targets:
            info[access_targets[target]] |= levels

    return info

def make_I(hier=DEFAULT_ACCESS_HIER, targets=DEFAULT_ACCESS_TARGETS):
    """
    Returns a partial that makes using `I` more convenient.
    """
    return partial(I, hier, targets)

class ModelColumnInfo(object):
    """
    Class that holds the information about model's column.
    """
    def __init__(self, model, name, db_column=True, permissions=None):
        self.model = model
        self.name = name
        self.db_column = db_column
        self.permissions = permissions if permissions is not None else {}

    #def get_value(self):
    #    return getattr(self.model, self.name)
    #def set_value(self, value):
    #    setattr(self.model, self.name, value)
    #value = property(get_value, set_value)



import string

def is_printable(value):
    """
    Returns True if value is a string, containing only printable characters.
    Returns False otherwise.
    """
    return isinstance(value, basestring) \
           and all(c in string.printable for c in value)



"""
When we think about HTTP API, one point is in what format(s) we exchange data.
I believe hardcoding is not a viable option and formats have to be configurable
and extensible.

In ToyBox, there are two concepts for this, called *serializers* and
*deserializers*. The names are hopefully self-describing.
"""

import json
import collections
import decimal

class ExtendedJSONEncoder(json.JSONEncoder):
    """
    Extended JSON encoder, that may come useful when implementing web APIs.

    Features for top-level objects:

    - If object has `to_json` attribute, it is used to represent object.
      Return string representation from there, and it'll get straight to the
      resulting JSON document.

    - If object is a namedtuple (or anything similar that's a tuple and
      has `_fields` attribute), it is represented as an JSON object (dict).

    Features for any objects, including deeply contained ones:

    - If object has `as_dict` or `as_list` methods, they're used to represent
      it as JSON object or list, respectively.

    - If object has `isoformat` method (for example, a `datetime` does) it's
      represented as JSON string.

    - `decimal.Decimal` values are represented as JSON strings (not numbers).

    - Otherwise, standard `json.JSONEncoder` rules apply.

    Usage example:

    >>> json.dumps(collections.namedtuple("whatever", "spam eggs")(
    ...     datetime.datetime(2012, 1, 1, 0, 0, 0), decimal.Decimal("42.42")),
    ...     cls=ExtendedJSONEncoder)
    '{"spam": "2012-01-01T00:00:00", "eggs": "42,42"}'

    """
    def encode(self, obj):
        if hasattr(obj, "to_json"):
            return obj.to_json()
        if isinstance(obj, tuple) and hasattr(obj, "_fields"):
            obj = collections.OrderedDict((k, v) for k, v in zip(obj._fields, obj))
        return super(ExtendedJSONEncoder, self).encode(obj)

    def default(self, obj):
        if hasattr(obj, "as_dict"):
            return obj.as_dict()
        elif hasattr(obj, "as_list"):
            return obj.as_list()
        elif hasattr(obj, "isoformat"):
            return obj.isoformat()
        elif isinstance(obj, decimal.Decimal):
            return str(obj)
        return super(ExtendedJSONEncoder, self).default(obj) # pragma: no cover

class JSON(object):
    """
    JSON serializer and deserializer.
    """
    mime_types = [
        "application/json",          # RFC4627, rest are compatibility
        "application/x-javascript",
        "text/javascript",
        "text/x-javascript",
        "text/x-json",
    ]

    @staticmethod
    def serialize(data):
        return json.dumps(data, cls=ExtendedJSONEncoder)

    @staticmethod
    def deserialize(data):
        return json.loads(data) # pragma: no cover



#!/usr/bin/env python

from setuptools import setup

setup(
    name="Flask-ToyBox",
    version="0.0.1",
    url="https://github.com/drdaeman/flask-toybox/",
    license="MIT",
    author="Aleksey Zhukov",
    author_email="drdaeman@drdaeman.pp.ru",
    description="Create somehow-RESTful HTTP APIs with Flask",
    long_description=open("README.rst", "r").read(),
    packages=["flask_toybox"],
    test_suite="tests",
    zip_safe=False,
    platforms="any",
    install_requires=["Flask >=0.9"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Web Environment",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)


import unittest

from flask.ext.toybox.sqlalchemy import SAModelMixin, saModelView
from flask.ext.toybox.permissions import make_I
from flask.ext.toybox import ToyBox
from flask import Flask, g
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String
import json

Base = declarative_base()

I = make_I()
class User(Base, SAModelMixin):
    __tablename__ = "test_users"

    id = Column(Integer, primary_key=True)
    username = Column(String, info=I("r:all,w:none"))
    fullname = Column(String, info=I("rw:all"))

    def __init__(self, username, fullname):
        self.username = username
        self.fullname = fullname

    def __repr__(self):
        return "<{0}: {1}, {2}>".format(self.__class__.__name__,
                                        self.username, self.fullname)

class SQLAlchemyModelTestCase(unittest.TestCase):
    def setUp(self):
        # Set up SQLAlchemy models
        engine = create_engine('sqlite:///:memory:', echo=False)
        Base.metadata.create_all(engine)
        Session = scoped_session(sessionmaker(bind=engine))

        # Create some models
        db_session = Session()
        db_session.add(User("spam", "Spam"))
        db_session.add(User("ham", "Ham"))
        db_session.add(User("eggs", "Eggs"))
        db_session.commit()

        # Set up Flask
        app = Flask(__name__)
        app.debug = True

        # Set up ToyBox
        toybox = ToyBox(app)

        class UserView(saModelView(db_session)):
            model = User

            #def fetch_object(self, username):
            #    obj = db_session.query(User).filter_by(username=username).one()
            #    g.etagger.set_object(obj)
            #    return obj
            def save_object(self, obj):
                # In a real code, this should be done by a middleware/wrapper.
                # However, this is a test, so we simplify things a bit.
                # Don't commit from here in production!
                db_session.commit()
        app.add_url_rule("/users/<username>", view_func=UserView.as_view("user"))

        self.app = app.test_client()

    def test_get(self):
        response = self.app.get("/users/spam", headers={"Accept": "application/json"})
        self.assertEqual(response.status_code, 200, response.status)

        reference = {"username": "spam", "fullname": "Spam"}
        data = json.loads(response.data)
        for k, v in reference.items():
            self.assertEqual(data[k], v)

        etag = response.headers.get("ETag", None)
        self.assertIsNotNone(etag)
        response = self.app.get("/users/spam", headers={
            "Accept": "application/json",
            "If-None-Match": etag
        })
        self.assertEqual(response.status_code, 304, response.status)

    def test_patch(self):
        response = self.app.get("/users/eggs",
                                headers={"Accept": "application/json"})
        self.assertEqual(response.status_code, 200)

        data = json.loads(response.data)
        self.assertIn("username", data)
        self.assertIn("fullname", data)
        self.assertEqual(data["username"], "eggs")

        etag = response.headers.get("ETag", None)
        self.assertIsNotNone(etag)

        response = self.app.patch(
            "/users/eggs",
            headers={"Accept": "application/json",},
            data=json.dumps({"fullname": "Python Eggs"}),
            content_type="application/json"
        )
        self.assertEqual(response.status_code, 428, response.status)

        response = self.app.patch(
            "/users/eggs",
            headers={
                "Accept": "application/json",
                "If-Match": etag
            },
            data = json.dumps({"fullname": "Python Eggs"}),
            content_type="application/json"
        )
        self.assertEqual(response.status_code, 204, response.data)
        response = self.app.get("/users/eggs",
                                headers={"Accept": "application/json"})
        self.assertEqual(response.status_code, 200, response.status)
        data = json.loads(response.data)
        self.assertIn("username", data)
        self.assertIn("fullname", data)
        self.assertEqual(data["fullname"], "Python Eggs")

    def test_patch_non_writeable(self):
        response = self.app.get("/users/eggs",
                                headers={"Accept": "application/json"})
        self.assertEqual(response.status_code, 200)

        etag = response.headers.get("ETag", None)
        self.assertIsNotNone(etag)
        response = self.app.patch(
            "/users/eggs",
            headers={
                "Accept": "application/json",
                "If-Match": etag
            },
            data = json.dumps({"username": "eggs2"}),
            content_type="application/json"
        )
        self.assertEqual(response.status_code, 422, response.data)

        response = self.app.get("/users/eggs",
                                headers={"Accept": "application/json"})
        self.assertEqual(response.status_code, 200, response.status)